#include "include/type.h"
#include "include/stddef.h"
#include "include/printk.h"
#include "include/panic.h"
#include "include/page.h"

#define MAX_ORDER 10 // 最大块大小为 2^10 * PAGE_SIZE

typedef struct free_area {
    struct free_area *next;
} free_area_t;

static free_area_t free_areas[MAX_ORDER + 1];
static uint8_t *memory_start;
static size_t total_pages;

void buddy_init(uint8_t *mem_start, size_t num_pages) {
    memory_start = mem_start;
    total_pages = num_pages;

    for (int i = 0; i <= MAX_ORDER; i++) {
        free_areas[i].next = NULL;
    }

    // 初始化整个内存为一个大块
    free_page_buddy(mem_start, MAX_ORDER);
}

void *alloc_page_buddy(size_t order) {
    if (order > MAX_ORDER) {
        panic("Order too large!");
        return NULL;
    }

    for (int i = order; i <= MAX_ORDER; i++) {
        if (free_areas[i].next) {
            free_area_t *block = free_areas[i].next;
            free_areas[i].next = block->next;

            // 将大块分割成小块
            for (int j = i - 1; j >= order; j--) {
                uint8_t *buddy = (uint8_t *)block + (1 << j) * PAGE_SIZE;
                free_page_buddy(buddy, j);
            }
            return block;
        }
    }
    panic("Out of memory!");
    return NULL;
}

void free_page_buddy(void *page, size_t order) {
    if (!page || order > MAX_ORDER) return;
    free_area_t *block = (free_area_t *)page;
    block->next = free_areas[order].next;
    free_areas[order].next = block;

    // 尝试合并伙伴块
    uint64_t addr = (uint64_t)page - (uint64_t)memory_start;
    uint64_t buddy_addr = addr ^ ((uint64_t)1 << (order + PAGE_SHIFT));
    uint8_t *buddy = (uint8_t *)memory_start + buddy_addr;

    free_area_t **prev = &free_areas[order].next;
    while (*prev) {
        if (*prev == (free_area_t *)buddy) {
            *prev = (*prev)->next;
            free_page_buddy((void *)((addr < buddy_addr) ? addr : buddy_addr), order + 1);
            return;
        }
        prev = &(*prev)->next;
    }
}