#include "include/printk.h"

int printk(const char *format, ...) {
    va_list args;
    va_start(args, format);
    int ret = vprintk(format, args);
    va_end(args);
    return ret;
}

int vprintk(const char *format, va_list args) {
    char buf[PRINTK_BUF_SIZE];
    int ret = vsprintk(buf, format, args);
    for (int i = 0; buf[i] != '\0'; i++) {
        putchar(buf[i]);
    }
    return ret;
}

int vsprintk(char *out, const char *format, va_list args) {
    return vsnprintk(out, PRINTK_BUF_SIZE, format, args);
}

void i2str(int val, char *buf, int base, bool upper) {
    static const char digits[] = "0123456789abcdef";
    static const char digits_upper[] = "0123456789ABCDEF";
    bool negative = false;
    int i = 0;
    if (val < 0 && base == 10) {
        negative = true;
        val = -val;
    }
    do {
        buf[i++] = upper ? digits_upper[val % base] : digits[val % base];
        val /= base;
    } while (val);
    if (negative) {
        buf[i++] = '-';
    }
    buf[i] = '\0';
    for (int j = 0; j < i / 2; j++) {
        char tmp = buf[j];
        buf[j] = buf[i - j - 1];
        buf[i - j - 1] = tmp;
    }
}

void ul2str(unsigned long val, char *buf, int base, bool upper) {
    static const char digits[] = "0123456789abcdef";
    static const char digits_upper[] = "0123456789ABCDEF";
    int i = 0;
    do {
        buf[i++] = upper ? digits_upper[val % base] : digits[val % base];
        val /= base;
    } while (val);
    buf[i] = '\0';
    for (int j = 0; j < i / 2; j++) {
        char tmp = buf[j];
        buf[j] = buf[i - j - 1];
        buf[i - j - 1] = tmp;
    }
}

int vsnprintk(char *out, size_t n, const char *format, va_list args) {
    const char* p = format;
    char* buf = out;
    while (*p != '\0') {
        if (*p != '%') {
            *buf++ = *p++;
            continue;
        }
        p++;
        
        bool is_long = false;
        if (*p == 'l') {
            is_long = true;
            p++;
        }

        switch (*p) {
            case 'd': {
                if (is_long) {
                    long val = va_arg(args, long);
                    char num[32];
                    if (val < 0) {
                        *buf++ = '-';
                        val = -val;
                    }
                    ul2str((unsigned long)val, num, 10, false);
                    char *num_p = num;
                    while (*num_p != '\0') {
                        *buf++ = *num_p++;
                    }
                } else {
                    int val = va_arg(args, int);
                    char num[32];
                    i2str(val, num, 10, false);
                    char *num_p = num;
                    while (*num_p != '\0') {
                        *buf++ = *num_p++;
                    }
                }
                p++;
                break;
            }
            case 'u': {
                if (is_long) {
                    unsigned long val = va_arg(args, unsigned long);
                    char num[32];
                    ul2str(val, num, 10, false);
                    char *num_p = num;
                    while (*num_p != '\0') {
                        *buf++ = *num_p++;
                    }
                } 
                else {
                    unsigned int val = va_arg(args, unsigned int);
                    char num[32];
                    ul2str(val, num, 10, false);
                    char *num_p = num;
                    while (*num_p != '\0') {
                        *buf++ = *num_p++;
                    }
                }
                p++;
                break;
            }
            case 'x':
            case 'X': {
                bool upper = (*p == 'X');
                if (is_long) {
                    unsigned long val = va_arg(args, unsigned long);
                    char num[32];
                    ul2str(val, num, 16, upper);
                    char *num_p = num;
                    while (*num_p != '\0') {
                        *buf++ = *num_p++;
                    }
                } 
                else {
                    unsigned int val = va_arg(args, unsigned int);
                    char num[32];
                    i2str(val, num, 16, upper);
                    char *num_p = num;
                    while (*num_p != '\0') {
                        *buf++ = *num_p++;
                    }
                }
                p++;
                break;
            }
            case 'p': {
                void* ptr = va_arg(args, void*);
                *buf++ = '0';
                *buf++ = 'x';
                char num[32];
                ul2str((unsigned long)ptr, num, 16, false);
                char *num_p = num;
                while (*num_p != '\0') {
                    *buf++ = *num_p++;
                }
                p++;
                break;
            }
            case 's': {
                const char* str = va_arg(args, const char*);
                while (*str != '\0') {
                    *buf++ = *str++;
                }
                p++;
                break;
            }
            case '%': {
                *buf++ = '%';
                p++;
                break;
            }
            default: {
                *buf++ = *p++;
                break;
            }
        }
    }
    
    int ret = buf - out;
    if (ret < n) {
        *buf = '\0';
        return ret;
    }
    else {
        out[n - 1] = '\0';
        return n - 1;
    }
}
