#include "include/proc.h"
#include "include/page.h"
#include "include/string.h"
#include "include/kmallo.h"
#include "include/printk.h"


static int next_pid = 1;
static PCB proc_table[MAX_PROC_NUM];  
static PCB* current = NULL;           

inline int alloc_pid() {
    return next_pid++;
}

void init_proc() {
    for(int i = 0; i < MAX_PROC_NUM; i++) {
        proc_table[i].state = PROC_UNUSED;
        proc_table[i].pid = 0;
    }
}

PCB* alloc_pcb() {
    for(int i = 0; i < MAX_PROC_NUM; i++) {
        if(proc_table[i].state == PROC_UNUSED) {
            PCB* p = &proc_table[i];
            memset(p, 0, sizeof(PCB));
            
            p->state = PROC_EMBRYO;
            p->pid = alloc_pid();
            p->priority = DEFAULT_PRIORITY;
            p->kstack = (uint64_t*)alloc_page();
            if (!p->kstack) {
                goto fail;
            }
            p->stack = (uint64_t*)alloc_page();
            if (!p->stack) {
                goto fail;
            }
            
            p->kstack_va = PROC_KERNEL_STACK_BASE - (i << 
                                KERNEL_STACK_SHIFT);
            if (p->kstack_va <= (uint64_t)p->kstack) {
                goto fail;
            }
            
            p->pagetable = create_pagetable();
            if(!p->pagetable) {
                #ifdef DEBUG
                printk("  Failed to create page table!\n");
                #endif
                goto fail;
            }
            
            if(map_page(p->pagetable, 
                       USER_STACK_BASE,
                       (uint64_t)p->stack,
                       PTE_R | PTE_W | PTE_U) < 0) {
                #ifdef DEBUG
                printk("  Failed to map user stack!\n");
                #endif
                goto fail;
            }
                
            // map kernel stack
            if(map_page(p->pagetable,
                       p->kstack_va,
                       (uint64_t)p->kstack,
                       PTE_R | PTE_W) < 0) {
                #ifdef DEBUG
                printk("  Failed to map kernel stack!\n");
                #endif
                goto fail;
            }
                
            p->context = (Context*)(p->kstack_va + KERNEL_STACK_SIZE - 
                sizeof(Context));
            memset((void*)p->kstack, 0, PAGE_SIZE);
            return p;
            
        fail:
            if(p->stack) free_page(p->stack);
            if(p->kstack) free_page(p->kstack);
            if(p->pagetable) free_pagetable(p->pagetable, 2);
            memset(p, 0, sizeof(PCB));
            p->state = PROC_UNUSED;
            #ifdef DEBUG
            printk("PCB %d allocation failed\n", i);
            #endif
            return NULL;
        }
    }
    #ifdef DEBUG
    printk("No free PCB available!\n");
    #endif
    return NULL;
}

void schedule() {
    PCB* prev = current;
    PCB* next = NULL;

    while(1) {
        for(int i = 0; i < MAX_PROC_NUM; i++) {
            if(proc_table[i].state == PROC_RUNNABLE) {
                next = &proc_table[i];
                break;
            }
        }
        if(next) {
            next->state = PROC_RUNNING;
            if(prev != next) {
                Context** old_ctx = prev ? &prev->context : NULL;
                current = next;
                switch_context(old_ctx, next->context);
            }
            break;
        }
    }
}

void yield() {
    if(current) {
        current->state = PROC_RUNNABLE;
        schedule();
    }
}

void exit() {
    if(current) {
        current->state = PROC_ZOMBIE;
        free_pagetable(current->pagetable, 2);
        free_page(current->stack);
        free_page(current->kstack);
        // TODO: clean page table
        schedule();
    }
}

void fork() {
    PCB* child = alloc_pcb();
    if(!child) return;
    
    if(current) {
        child->parent = current;
        child->chan = 0;
        
        child->pagetable = copy_pagetable(current->pagetable);
        if (!child->pagetable) {
            free_page(child->stack);
            free_page(child->kstack);
            child->state = PROC_UNUSED;
            return;
        }
        
        if(map_page(child->pagetable,
                   USER_STACK_BASE,
                   (uint64_t)child->stack,
                   PTE_R | PTE_W | PTE_U) < 0) {
            free_pagetable(child->pagetable, 2);
            free_page(child->stack);
            free_page(child->kstack);
            child->state = PROC_UNUSED;
            return;
        }
        
        memcpy(child->stack, current->stack, PAGE_SIZE);
        
        *(child->context) = *(current->context);
        child->state = PROC_RUNNABLE;
    }
}

void wait() {
    if(!current) return;
    
    for(int i = 0; i < MAX_PROC_NUM; i++) {
        if(proc_table[i].parent == current && 
           proc_table[i].state == PROC_ZOMBIE) {
            proc_table[i].state = PROC_UNUSED;
            return;
        }
    }
    
    current->state = PROC_SLEEPING;
    schedule();
}

void sleep(void* chan) {
    if(!current)
        return;
        
    current->chan = chan;
    current->state = PROC_SLEEPING;
    schedule();
}

void wakeup(void* chan) {
    for(int i = 0; i < MAX_PROC_NUM; i++) {
        if(proc_table[i].state == PROC_SLEEPING && 
           proc_table[i].chan == chan) {
            proc_table[i].state = PROC_RUNNABLE;
            proc_table[i].chan = 0;
        }
    }
}
